@@ -4,19 +4,24 @@ use 5.010;
 use strict;
 use warnings;
 
-our $VERSION = '0.07';
+our $VERSION = '0.08';
 
+use Cassandra::Client::AsyncAnyEvent;
+use Cassandra::Client::AsyncEV;
 use Cassandra::Client::Config;
 use Cassandra::Client::Connection;
-use Cassandra::Client::Util qw/series/;
-use Cassandra::Client::AsyncIO;
-use Cassandra::Client::AsyncAnyEvent;
 use Cassandra::Client::Metadata;
+use Cassandra::Client::Policy::Queue::Default;
+use Cassandra::Client::Policy::Retry::Default;
+use Cassandra::Client::Policy::Retry;
+use Cassandra::Client::Policy::Throttle::Adaptive;
 use Cassandra::Client::Pool;
-use Cassandra::Client::AdaptiveThrottler;
+use Cassandra::Client::Util qw/series/;
+
+use Clone qw/clone/;
 use List::Util qw/shuffle/;
 use Promises qw/deferred/;
-use Clone qw/clone/;
+use Time::HiRes ();
 
 sub new {
     my ($class, %args)= @_;
@@ -26,12 +31,14 @@ sub new {
         connect_callbacks => undef,
         shutdown          => 0,
         shutdown_cb       => undef,
+
+        active_queries    => 0,
     }, $class;
 
     my $options= Cassandra::Client::Config->new(
         \%args
     );
-    my $async_class= $options->{anyevent} ? "Cassandra::Client::AsyncAnyEvent" : "Cassandra::Client::AsyncIO";
+    my $async_class= $options->{anyevent} ? "Cassandra::Client::AsyncAnyEvent" : "Cassandra::Client::AsyncEV";
     my $async_io= $async_class->new(
         options => $options,
     );
@@ -43,13 +50,20 @@ sub new {
         options  => $options,
         metadata => $metadata,
     );
+    my $command_queue= Cassandra::Client::Policy::Queue::Default->new(
+        %{ $options->{command_queue_config} || {} },
+    );
+    my $retry_policy= Cassandra::Client::Policy::Retry::Default->new();
 
     $self->{options}= $options;
     $self->{async_io}= $async_io;
     $self->{metadata}= $metadata;
     $self->{pool}= $pool;
+    $self->{command_queue}= $command_queue;
+    $self->{retry_policy}= $retry_policy;
     if ($options->{throttler}) {
-        my $throttler_class= "Cassandra::Client::$options->{throttler}";
+        $options->{throttler}= 'Adaptive' if $options->{throttler} eq 'AdaptiveThrottler'; # Temporary.
+        my $throttler_class= "Cassandra::Client::Policy::Throttle::$options->{throttler}";
         $options->{throttler}= $throttler_class->new(%{$options->{throttler_config}});
     } else {
         $self->{throttler}= undef;
@@ -210,37 +224,50 @@ sub _wait_for_schema_agreement {
 sub _command {
     my ($self, $command, $callback, $args)= @_;
 
+    my $command_info= {
+        start_time => Time::HiRes::time(),
+    };
+
     # Handle overloads
     goto FAILFAST if $self->{throttler} && $self->{throttler}->should_fail();
 
+    goto OVERFLOW if $self->{active_queries} >= $self->{options}{max_concurrent_queries};
+
     goto SLOWPATH if !$self->{connected};
 
     my $connection= $self->{pool}->get_one;
     goto SLOWPATH if !$connection;
 
+    $self->{active_queries}++;
     $connection->$command(sub {
         my ($error, $result)= @_;
         if (my $throttler= $self->{throttler}) {
             $throttler->count($error);
         }
 
-        if ($error && ref($error) && $error->retryable) {
-            return $self->_command_slowpath($command, $callback, $args);
-        }
+        $self->{active_queries}--;
+        $self->_command_dequeue if $self->{command_queue}{has_any};
+
+        return $self->_command_failed($command, $callback, $args, $command_info, $error) if $error;
         return _cb($callback, $error, $result);
     }, @$args);
 
     return;
 
 SLOWPATH:
-    return $self->_command_slowpath($command, $callback, $args);
+    return $self->_command_slowpath($command, $callback, $args, $command_info);
 
 FAILFAST:
     return _cb($callback, "Client-induced failure by throttling mechanism");
+
+OVERFLOW:
+    return $self->_command_enqueue($command, $callback, $args, $command_info);
 }
 
 sub _command_slowpath {
-    my ($self, $command, $callback, $args)= @_;
+    my ($self, $command, $callback, $args, $command_info)= @_;
+
+    $self->{active_queries}++;
 
     series([
         sub {
@@ -251,6 +278,12 @@ sub _command_slowpath {
             $self->{pool}->get_one_cb($next);
         }, sub {
             my ($next, $connection)= @_;
+            # Yes, if we immediately take the slow path, which we would if we're not connected, we're going to throttle twice
+            # For now, I'm okay with that, but let's mark it with a TODO anyway.
+            # XXX
+            if ($self->{throttler} && $self->{throttler}->should_fail()) {
+                return $next->("Client-induced failure by throttling mechanism");
+            }
             $connection->$command($next, @$args);
         }
     ], sub {
@@ -258,11 +291,71 @@ sub _command_slowpath {
         if (my $throttler= $self->{throttler}) {
             $throttler->count($error);
         }
+
+        $self->{active_queries}--;
+        $self->_command_dequeue if $self->{command_queue}{has_any};
+
+        return $self->_command_failed($command, $callback, $args, $command_info, $error) if $error;
         return _cb($callback, $error, $result);
     });
     return;
 }
 
+sub _command_retry {
+    my ($self, $command, $callback, $args, $command_info)= @_;
+
+    $command_info->{retries}++;
+
+    my $delay= 0.1 * (2 ** $command_info->{retries});
+    $self->{async_io}->timer(sub {
+        if ($self->{active_queries} >= $self->{options}{max_concurrent_queries}) {
+            $self->_command_enqueue($command, $callback, $args, $command_info);
+        } else {
+            $self->_command_slowpath($command, $callback, $args, $command_info);
+        }
+    }, $delay);
+}
+
+sub _command_failed {
+    my ($self, $command, $callback, $args, $command_info, $error)= @_;
+
+    return $callback->($error) unless ref $error;
+
+    my $retry_decision;
+    if ($error->{do_retry}) {
+        $retry_decision= Cassandra::Client::Policy::Retry::retry;
+    } elsif ($error->code == 0x1100) {
+        $retry_decision= $self->{retry_policy}->on_write_timeout(undef, @$error{qw/cl write_type blockfor received/}, ($command_info->{retries}||0));
+    } elsif ($error->code == 0x1200) {
+        $retry_decision= $self->{retry_policy}->on_read_timeout(undef, @$error{qw/cl blockfor received data_retrieved/}, ($command_info->{retries}||0));
+    } elsif ($error->code == 0x1000) {
+        $retry_decision= $self->{retry_policy}->on_unavailable(undef, @$error{qw/cl required alive/}, ($command_info->{retries}||0));
+    } else {
+        $retry_decision= Cassandra::Client::Policy::Retry::rethrow;
+    }
+
+    if ($retry_decision && $retry_decision eq 'retry') {
+        return $self->_command_retry($command, $callback, $args, $command_info);
+    }
+
+    return $callback->($error);
+}
+
+sub _command_enqueue {
+    my ($self, $command, $callback, $args, $command_info)= @_;
+    if (my $error= $self->{command_queue}->enqueue([$command, $callback, $args, $command_info])) {
+        return $callback->("Cannot $command: $error");
+    }
+    return;
+}
+
+sub _command_dequeue {
+    my ($self)= @_;
+    my $item= $self->{command_queue}->dequeue or return;
+    $self->_command_slowpath(@$item);
+    return;
+}
+
 # Utility functions that wrap query functions
 sub _each_page {
     my ($self, $callback, $query, $params, $attribs, $page_callback)= @_;
@@ -295,6 +388,21 @@ sub _each_page {
 
 
 # Utility functions for callers
+sub _get_stacktrace {
+    # This gets called a lot. Let's keep it fast.
+
+    my $trace= '';
+    my ($c, $file, $line)= caller(1);
+    $trace .= "    $c ($file:$line)\n";
+    ($c, $file, $line)= caller(2) or goto DONE;
+    $trace .= "    $c ($file:$line)\n";
+    ($c, $file, $line)= caller(3) or goto DONE;
+    $trace .= "    $c ($file:$line)\n";
+
+DONE:
+    return $trace;
+}
+
 sub _cb {
     my $cb= shift;
     eval {
@@ -331,12 +439,13 @@ sub _mkpromise {
     my ($sub)= @_;
     return sub {
         my $self= shift;
+        my $trace= &_get_stacktrace;
         my $deferred= deferred;
 
         $sub->($self, sub {
             my ($error, @output)= @_;
             if ($error) {
-                $deferred->reject($error);
+                $deferred->reject("$error\n\nTrace:\n$trace");
             } else {
                 $deferred->resolve(@output);
             }
@@ -350,10 +459,11 @@ sub _mkfuture {
     my ($sub)= @_;
     return sub {
         my $self= shift;
+        my $trace= &_get_stacktrace;
         $sub->($self, $self->{async_io}->wait(my $w), @_);
         return sub {
             my ($error, @output)= $w->();
-            if ($error) { die $error; }
+            if ($error) { die "$error\n\nTrace:\n$trace"; }
             return @output;
         };
     }
@@ -522,7 +632,7 @@ Disconnect all connections and abort all current queries. After this, the C<Cass
 
 =item $client->wait_for_schema_agreement()
 
-I<Not implemented yet, this just does a C<sleep 1>!> Wait until all nodes agree on the schema version. Useful after changing table or keyspace definitions.
+Wait until all nodes agree on the schema version. Useful after changing table or keyspace definitions.
 
 =back
 
