@@ -19,6 +19,7 @@ use Cassandra::Client::Protocol qw/
     pack_shortbytes
     pack_stringmap
     pack_stringlist
+    unpack_errordata
     unpack_inet
     unpack_int
     unpack_metadata
@@ -79,7 +80,7 @@ sub get_local_status {
     series([
         sub {
             my ($next)= @_;
-            $self->execute_prepared($next, \"select key, data_center, host_id, broadcast_address, rack, release_version, tokens from system.local", undef, { consistency => 'local_one' });
+            $self->execute_prepared($next, \"select key, data_center, host_id, broadcast_address, rack, release_version, tokens, schema_version from system.local", undef, { consistency => 'local_one' });
         },
         sub {
             my ($next, $result)= @_;
@@ -92,6 +93,7 @@ sub get_local_status {
                 rack => $_->[4],
                 release_version => $_->[5],
                 tokens => $_->[6],
+                schema_version => $_->[7],
             } } @{$result->rows};
 
             $next->(undef, \%local_status);
@@ -107,7 +109,7 @@ sub get_peers_status {
     series([
         sub {
             my ($next)= @_;
-            $self->execute_prepared($next, \"select peer, data_center, host_id, preferred_ip, rack, release_version, tokens from system.peers", undef, { consistency => 'local_one' });
+            $self->execute_prepared($next, \"select peer, data_center, host_id, preferred_ip, rack, release_version, tokens, schema_version from system.peers", undef, { consistency => 'local_one' });
         },
         sub {
             my ($next, $result)= @_;
@@ -120,6 +122,7 @@ sub get_peers_status {
                 rack => $_->[4],
                 release_version => $_->[5],
                 tokens => $_->[6],
+                schema_version => $_->[7],
             } } @{$result->rows};
 
             $next->(undef, \%network_status);
@@ -410,7 +413,10 @@ sub decode_result {
         return $callback->();
 
     } elsif ($result_type == RESULT_SCHEMA_CHANGE) { # Schema change
-        return $self->wait_for_schema_agreement($callback);
+        return $self->wait_for_schema_agreement(sub {
+            # We may be passed an error. Ignore it, our query succeeded
+            $callback->();
+        });
 
     } else {
         return $callback->("Query executed successfully but got an unexpected response type");
@@ -420,8 +426,41 @@ sub decode_result {
 
 sub wait_for_schema_agreement {
     my ($self, $callback)= @_;
-    sleep 1; # I am very sorry.
-    return $callback->();
+
+    my $waited= 0;
+    my $wait_delay= 0.5;
+    my $max_wait= 5;
+
+    my $wait_and_check; $wait_and_check= sub {
+        $waited += $wait_delay;
+        series([
+            sub {
+                my ($next)= @_;
+                $self->{async_io}->timer($next, $wait_delay);
+            },
+            sub {
+                my ($next)= @_;
+                $self->get_network_status($next);
+            },
+        ], sub {
+            my ($error, $network_status)= @_;
+            return $callback->($error) if $error;
+
+            my %versions;
+            $versions{$_->{schema_version}}= 1 for values %$network_status;
+            if (keys %versions > 1) {
+                if ($waited < $max_wait) {
+                    return $wait_and_check->();
+                } else {
+                    return $callback->("wait_for_schema_agreement timed out after $waited seconds");
+                }
+            }
+            return $callback->();
+        });
+    };
+
+    $wait_and_check->();
+    return;
 }
 
 
@@ -674,8 +713,19 @@ sub request {
             $self->{pending_write}= $data;
             $self->{async_io}->register_write($self->{fileno});
         } else {
+            # Oh, we failed to send out the request. That's bad. Let's first find out what happened.
             my $error= $!;
+
+            # We never actually sent our request, so take it out again
+            my $my_stream= delete $pending->{$stream_id};
+
             $self->shutdown(undef, $error);
+
+            # Now fail our stream properly, but include the retry notice
+            $my_stream->[0]->(Cassandra::Client::Error->new(
+                message  => "Disconnected: $error",
+                do_retry => 1,
+            ));
         }
     }
 
@@ -729,14 +779,11 @@ READ_NEXT:
                 } # Else: totally fine
 
             } elsif ($opcode == OPCODE_ERROR) {
-                my ($code, $message)= unpack('Nn/a', $body);
                 my ($cb, $dl)= @$stream_cb;
                 $$dl= 1;
-                $cb->(Cassandra::Client::Error->new(
-                    code => $code,
-                    message => $message,
-                    our_fault => ( $code != 0x1000 && $code != 0x1001 && $code != 0x1002 && $code != 0x1003 && $code != 0x1100 && $code != 0x1200 ),
-                ));
+
+                my $error= unpack_errordata($body);
+                $cb->($error);
 
             } else {
                 my ($cb, $dl)= @$stream_cb;
@@ -788,7 +835,7 @@ sub can_timeout {
     my ($self, $id)= @_;
     my $stream= delete $self->{pending_streams}{$id};
     $self->{pending_streams}{$id}= [ sub{}, \(my $zero= 0) ]; # fake it
-    $stream->[0]->(Cassandra::Client::Error->new(message => "Request timed out", our_fault => 0));
+    $stream->[0]->(Cassandra::Client::Error->new(message => "Request timed out", is_timeout => 1));
     return;
 }
 
@@ -816,7 +863,6 @@ sub shutdown {
     for (values %$pending) {
         $_->[0]->(Cassandra::Client::Error->new(
             message   => "Disconnected: $shutdown_reason",
-            our_fault => 0,
         ));
     }
 
